#!/usr/bin/env python
import os
import numpy as np
import cv2
import torch
from network.ulsd import ULSD
import torchvision.transforms.transforms as tfs
from PIL import Image
from config.cfg import parse
from scipy.special import comb
import subprocess
import iio

def normalized_read(img_path, cfg):
    image = Image.open(img_path).convert('RGB')
    image = image.resize(cfg.image_size, Image.BILINEAR)
    image = tfs.ToTensor()(image)
    image = tfs.Normalize(cfg.mean, cfg.std)(image)
    return image.unsqueeze(0)

# line = Nx2 numpy array, N corresponding to the numper of points to fit
# Adapted from https://stackoverflow.com/questions/12643079/b%C3%A9zier-curve-fitting-with-scipy
def bezier_approximation(line, order=3):
    def bpoly(n, t, k):
        """ Bernstein polynomial when a = 0 and b = 1. """
        return t ** k * (1 - t) ** (n - k) * comb(n, k)

    def bmatrix(T):
        """ Bernstein matrix for Bézier curves. """
        return np.matrix([[bpoly(order, t, k) for k in range(order + 1)] for t in T])

    def least_square_fit(points, M):
        M_ = np.linalg.pinv(M)
        return M_ * points

    ts = np.linspace(0, 1, line.shape[0])
    M = bmatrix(ts)
    return least_square_fit(line, M)

if __name__ == '__main__':
    # Parameter
    cfg = parse()

    img_path = cfg.img_path
    out = cfg.out
    print(img_path, cfg.model_path, cfg.score_thresh, cfg.out)

    # Load model
    device = 'cpu'
    model = ULSD(cfg).to(device)
    model_filename = os.path.join(cfg.model_path, cfg.model_name)
    checkpoint = torch.load(model_filename, map_location=device)
    if 'model' in checkpoint.keys():
        state_dict = checkpoint['model']
    else:
        state_dict = checkpoint
    model.load_state_dict(state_dict)
    model.eval()

    imgn = normalized_read(img_path, cfg)
    img = iio.read(img_path)
    h, w, c = img.shape

    with torch.no_grad():
        jmaps, joffs, line_preds, line_scores = model(imgn)
    line_preds = line_preds[0].numpy()
    line_scores = line_scores[0].numpy()
    # Only keep those above the threshold
    line_preds = line_preds[line_scores > cfg.score_thresh]
    line_scores = line_scores[line_scores > cfg.score_thresh]

    sx, sy = w / cfg.heatmap_size[0], h / cfg.heatmap_size[1]
    line_preds[..., 0] *= sx
    line_preds[..., 1] *= sy
        
    ##### OUTPUT ######
    out_svg = out + '.svg'
    out_txt = out + '.txt'
    out_png = out + '.png'
    out_eps = out + '.eps'


    ftxt = open(out_txt, "w")
    fcsv = open(out_svg, "w")
    feps = open(out_eps, "w")

    fcsv.write('<?xml version="1.0" standalone="no"?>\n<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"\n"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd"><svg version="1.1" xmlns="http://www.w3.org/2000/svg"\nxmlns:xlink="http://www.w3.org/1999/xlink"\nwidth="'+str(w)+'" height="'+str(h)+'" viewBox="0 0 '+str(w)+' '+str(h)+'">')

    feps.write("%!PS-Adobe-3.0 EPSF-3.0\n");
    feps.write(f"%%BoundingBox: 0 0 {w} {h}\n");
    feps.write("%%Creator: LSD, Line Segment Detector\n");
    feps.write(f"%%Title: ({out_eps})\n");
    feps.write("%%EndComments\n");

    color='black'
    width=1.5
    for tp_id, (line, score) in enumerate(zip(line_preds, line_scores)):
        # Cubic Bezier approximation, original code considers order 4 but these are not compatible with SVG.
        bline = bezier_approximation(line)

        fcsv.write(('<path d="M {x1} {y1} C {x2} {y2}, {x3} {y3}, {x4} {y4}" stroke="'+color+'" stroke-width="'+str(width)+'" fill="transparent"/>\n').format(x1=bline[0,0],y1=bline[0,1],x2=bline[1,0],y2=bline[1,1],x3=bline[2,0],y3=bline[2,1],x4=bline[3,0],y4=bline[3,1]))
        ftxt.write(('{x1} {y1} {x2} {y2} {x3} {y3} {x4} {y4} {score}\n').format(x1=bline[0,0],y1=bline[0,1],x2=bline[1,0],y2=bline[1,1],x3=bline[2,0],y3=bline[2,1],x4=bline[3,0],y4=bline[3,1],score=score) )
        feps.write("newpath {x1} {y1} moveto {x2} {y2} {x3} {y3} {x4} {y4} curveto 1.5 setlinewidth stroke\n".format(x1=bline[0,0],y1=h-bline[0,1],x2=bline[1,0],y2=h-bline[1,1],x3=bline[2,0],y3=h-bline[2,1],x4=bline[3,0],y4=h-bline[3,1]))
    fcsv.write('\n</svg>')
    feps.write("showpage\n");
    feps.write("%%EOF\n");

    ftxt.close() 
    fcsv.close() 
    feps.close()
    
    p2 = ['gs', '-dNOPAUSE', '-dBATCH', '-sDEVICE=pnggray', '-dGraphicsAlphaBits=4', '-r72', \
            '-dEPSCrop', '-sOutputFile='+out_png, out_eps]
    val1 = subprocess.run(p2)
    im = iio.read(out_png)
    im = im.mean(axis=-1)
    im = 255-im
    iio.write(out_png, im)

